c-------------------------------------------------------------------------------
c Program:    2D_Potential_FMM - A fast multipole (FMM) boundary element 
c             method (BEM) code for analyzing large-scale, general 2-D potential 
c             problems (governed by Laplace equation) using constant elements.
c
c Developers: Dr. Naoshi Nishimura at Kyoto University, Japan;
c             Dr. Yijun Liu at the University of Cincinnati, Cincinnati, OH.
c
c Version:    V.1.20.
c Released:   October 1, 2008.
c
c Copyright(c)2004-2009 By Kyoto University and University of Cincinnati.
c             This code is intended for educational use only. No part of 
c             the code can be used for any commercial applications/distributions
c             without prior written permissions of the original developers.
c
c-------------------------------------------------------------------------------

      implicit real*8(a-h,o-z)

	integer, allocatable    :: ia(:)
      complex*16, allocatable :: am(:)

	character*80 Prob_Title

      call CPU_Time(time0)

      open (4, file='input.fmm',        status='old')
      open (5, file='input.dat',        status='old')
      open (3, file='output.dat',       status='unknown')
      open (7, file='phi_boundary.plt', status='unknown')
      open (8, file='xy.plt',           status='unknown')
      open (9, file='phi_domain.plt',   status='unknown')

c Input the parameters

      read(4,*)        maxl, levmx, nexp, ntylr, tolerance 
      read(4,*)        maxia, ncellmx, nleafmx, mxl, nwksz
	read(5,'(a80)')  Prob_Title
	read(5,*)        n, nfield
	write(3,'(a80)') Prob_Title
	write(*,'(a80)') Prob_Title

c Estimate the maximum numbers of the cells and leaves, 
c and size of the pre-conditioning matrix, etc.

      if(ncellmx.le.0) ncellmx = max(4*n/maxl,100)
      if(nleafmx.le.0) nleafmx = max(ncellmx/2,100)
      if(nwksz.le.0)   nwksz   = maxl*maxl*nleafmx
	ligw  = 20
	lrgw  = 1+n*(mxl+6)+mxl*(mxl+3)
	iwksz = n+3*nleafmx+1

	allocate (ia(maxia))

c Load the addresses (pointers) associated with the locations of the 
c variables to be stored in the large array "am"

	call lpointer(lp, ln, maxia, ia, n, nexp, ntylr, ncellmx,
     &        levmx, ligw, lrgw, nwksz, iwksz, nfield,
     &        l_n,       l_x,       l_y,     l_node,  l_dnorm,
     &        l_bc,      l_a,       l_b,     l_xmax,
     &        l_xmin,    l_ymax,    l_ymin,  l_ielem, l_itree,
     &        l_level,   l_loct,    l_numt,  l_ifath, l_lowlev,
     &        l_maxl,    l_levmx,   l_nexp,  l_ntylr, l_tolerance,
     &        l_ncellmx, l_nleafmx, l_mxl,   l_u,     l_ax,
     &        l_sb,      l_sx,      l_ligw,  l_lrgw,  l_igwk,
     &        l_rgwk,    l_nwksz,   l_iwksz, l_rwork, l_iwork,
     &        l_xfield,  l_nfield,  l_f)

c Estimate the memory usage

      maxa = lp
      write(3,100) maxa*16/1.D6
      write(*,100) maxa*16/1.D6
100   format('  Memory size of the large block am  =', f12.1,' Mb'/)

c Allocate the large block 'am'

	allocate (am(maxa))

c Assign the parameters to the array am()

      call assigni(n,         am(l_n     ))
	call assigni(maxl,      am(l_maxl  ))
	call assigni(levmx,     am(l_levmx ))
	call assigni(nexp,      am(l_nexp  ))
	call assigni(ntylr,     am(l_ntylr ))
	call assignd(tolerance, am(l_tolerance))
	call assigni(ncellmx,   am(l_ncellmx))
	call assigni(nleafmx,   am(l_nleafmx))
	call assigni(mxl,       am(l_mxl   ))
	call assigni(ligw,      am(l_ligw  ))
	call assigni(lrgw,      am(l_lrgw  ))
	call assigni(nwksz,     am(l_nwksz ))
	call assigni(iwksz,     am(l_iwksz ))
	call assigni(nfield,    am(l_nfield))

c Call the FMM BEM main program

	call fmmmain(maxa,   maxia,         am,           ia,
     &      am(l_n),       am(l_x),       am(l_y),      am(l_node),
     &      am(l_dnorm),   am(l_bc),      am(l_a),      am(l_b), 
     &      am(l_xmax),    am(l_xmin),    am(l_ymax),   am(l_ymin),
     &      am(l_ielem),   am(l_itree),   am(l_level),  am(l_loct), 
     &      am(l_numt),    am(l_ifath),   am(l_lowlev), am(l_maxl), 
     &      am(l_levmx),   am(l_nexp),    am(l_ntylr),  am(l_tolerance),
     &      am(l_ncellmx), am(l_nleafmx), am(l_mxl),    am(l_u), 
     &      am(l_ax),      am(l_nfield),  am(l_xfield), am(l_f),
     &      am(l_sb),      am(l_sx),      am(l_igwk),   am(l_rgwk), 
     &      am(l_ligw),    am(l_lrgw),    am(l_nwksz),  am(l_iwksz),   
     &      am(l_rwork),   am(l_iwork))

c Estimate the total CPU time

	call CPU_Time(time)
	write(3,*)
	write(*,*)
      write(3,*) ' Total CPU time used =', time-time0, '(sec)'
      write(*,*) ' Total CPU time used =', time-time0, '(sec)'

      stop
      end

c-------------------------------------------------------------------------------
c Definition of Variables:
c
c maxa      = maximum size of the array am
c maxia     = maximum number of variables allowed
c am        = a large array storing the variables for the SLATEC GMRES solver
c ia        = an array storing the locations of the variables in the array am
c
c n         = number of elements (= number of nodes)
c x         = coordinates of the nodes
c y         = coordinates of the end points of the elements
c node      = element connectivity
c dnorm     = normal at each node
c bc        = BC type and value
c
c a         = multipole expansion moments
c b         = local expansion coefficients
c xmax,xmin = maximum and minimum x coordinate
c ymax,ymin = maximum and minimum y coordinate
c ielem     = ielem(i) gives the original element number for i-th element in 
c             the quad-tree structure
c itree     = itree(c) gives the cell location of c-th cell within each tree level
c loct      = elements included in the c-th cell are listed starting at 
c             the loct(c)-th place in the array ielem
c numt      = numt(c) gives the number of elements included in the c-th cell
c ifath     = ifath(c) gives the cell number of the parent cell of the c-th cell
c level     = level l cells start at the level(l)-th cell in the tree
c lowlev    = number of the tree levels
c
c maxl      = maximum number of elements allowed in a leaf
c levmx     = maximum number of levels allowed in the tree structure
c nexp      = number of terms in multipole expansion
c ntylr     = number of terms in local expansion
c tolerance = GMRES solution convergence tolerance
c ncellmx   = maximum number of cells allowed in the tree
c nleafmx   = maximum number of leaves allowed in the tree
c mxl       = maximum dimension of Krylov subspace (used in GMRES)
c
c u         = first stores b vector; then solution vector of the system Ax = b
c ax        = resulting vector of multiplication Ax
c nfield    = number of the field points inside the domain
c xfield    = coordinates of the field points inside the domain
c f         = values of the potential at the field points inside the domain
c
c The following variables and arrays are used in the SLATEC GMRES solver:
c   sb,sx,igwk,rgwk,ligw,lrgw,nwksz,iwksz,rwork,iwork
c
c-------------------------------------------------------------------------------